# coding:utf-8

from errno import EINVAL
from typing import Optional
from typing import Sequence

from xkits import add_command
from xkits import argp
from xkits import commands
from xkits import run_command

from .attribute import __urlhome__
from .attribute import __version__
from .password import Argon2Hasher

DEFAULT_FILE = ".pwhashed"


def get_password(password: Optional[str] = None) -> str:
    if password is None:
        password = input("password: ")
    return password


@add_command("verify", description="verify password")
def add_cmd_verify(_arg: argp):
    _arg.add_argument("--hash", dest="password_hash", metavar="FILE", default=DEFAULT_FILE,  # noqa:E501
                      help=f"encoded hash file, default is '{DEFAULT_FILE}'")
    _arg.add_argument(dest="password", metavar="TEXT", nargs="?",
                      help="password plaintext, default secure input via dialog")  # noqa:E501


@run_command(add_cmd_verify)
def run_cmd_verify(cmds: commands) -> int:
    password: str = get_password(cmds.args.password)
    password_hash: str = cmds.args.password_hash
    with open(password_hash, "r") as rhdl:
        hashed: str = rhdl.read().strip()
    if not Argon2Hasher(hashed).verify(password):
        cmds.stderr_red("password mismatch")
        return EINVAL
    cmds.stdout_green("password match")
    return 0


@add_command("encode", description="compute encoded hash")
def add_cmd_encode(_arg: argp):
    _arg.add_argument("--store", dest="password_hash", metavar="FILE", nargs="?", const=DEFAULT_FILE,  # noqa:E501
                      help=f"store encoded hash to file, default is '{DEFAULT_FILE}'")  # noqa:E501
    _arg.add_argument("--password", dest="password", metavar="TEXT",
                      help="password plaintext, default secure input via dialog")  # noqa:E501
    _arg.add_argument(dest="password_salt", metavar="SALT", nargs="?",
                      help="password salt, default is random value")


@run_command(add_cmd_encode)
def run_cmd_encode(cmds: commands) -> int:
    password: str = get_password(cmds.args.password)
    password_salt: Optional[str] = cmds.args.password_salt
    password_hash: Optional[str] = cmds.args.password_hash
    hasher: Argon2Hasher = Argon2Hasher.hash(
        password=password, salt=password_salt,
        time_cost=16, memory_cost=65536, parallelism=8,
        hash_len=64, salt_len=32)
    if isinstance(password_hash, str):
        with open(password_hash, "w") as whdl:
            cmds.stdout(f"store encoded hash to file: {password_hash}")
            whdl.write(hasher.hashed)
    else:
        cmds.stdout(hasher.hashed)
    return 0


@add_command("pwhasher", description="compute encoded hash or verify password")
def add_cmd(_arg: argp):
    pass


@run_command(add_cmd, add_cmd_encode, add_cmd_verify)
def run_cmd(cmds: commands) -> int:
    return 0


def main(argv: Optional[Sequence[str]] = None) -> int:
    cmds = commands()
    cmds.version = __version__
    return cmds.run(root=add_cmd, argv=argv, epilog=f"For more, please visit {__urlhome__}.")  # noqa:E501



# coding:utf-8

from os import urandom
from typing import Union

from argon2 import PasswordHasher
from argon2 import Type
from argon2.exceptions import VerifyMismatchError


class Secret():
    """Hashed password digest"""

    def __init__(self, key: str):
        self.__key: str = key

    def __hash__(self) -> int:
        return hash(self.key)

    def __repr__(self) -> str:
        return f"Secret({self.key})"

    def __str__(self) -> str:
        return self.key

    def __eq__(self, other: Union["Secret", str]) -> bool:
        return str(other) == self.key if isinstance(other, (Secret, str)) else False  # noqa:E501

    @property
    def key(self) -> str:
        """secret key"""
        return self.__key


class Salt():
    """Password salt"""
    MIN_LENGTH: int = 8  # minimum length
    DEF_LENGTH: int = 16  # default length

    def __init__(self, value: bytes):
        self.__value: bytes = value.ljust(self.MIN_LENGTH, b"x")

    @property
    def value(self) -> bytes:
        """salt value"""
        return self.__value

    @classmethod
    def format(cls, value: bytes, length: int = DEF_LENGTH) -> "Salt":
        """right-justified password salt"""
        return cls(value.rjust(length, b"x"))

    @classmethod
    def random(cls, length: int = DEF_LENGTH) -> "Salt":
        """generate random password salt"""
        return cls(urandom(length))

    @classmethod
    def generate(cls, value: Union[str, bytes, None] = None, length: int = DEF_LENGTH) -> "Salt":  # noqa:E501
        """generate password salt"""
        return cls.random(length) if value is None else cls.format(value.encode("utf-8") if isinstance(value, str) else value)  # noqa:E501


class Argon2Hasher():
    """Argon2 password hasher"""
    DEFAULT_TIME_COST: int = 8
    DEFAULT_MEMORY_COST: int = 65536
    DEFAULT_PARALLELISM: int = 4
    DEFAULT_HASH_LENGTH: int = 32
    DEFAULT_SALT_LENGTH: int = 16

    def __init__(self, hashed: str):
        self.__hashed: str = hashed
        if not isinstance(self.verify(__name__), bool):
            raise ValueError("Invalid hash")
        self.__secret: Secret = Secret(key=hashed.split("$")[-1])

    @property
    def hashed(self) -> str:
        """encoded hash"""
        return self.__hashed

    @property
    def secret(self) -> Secret:
        """secret key"""
        return self.__secret

    def verify(self, password: str) -> bool:
        """verify password is match"""
        try:
            return PasswordHasher().verify(self.hashed, password)
        except VerifyMismatchError:
            return False

    @classmethod
    def hash(cls, password: str, salt: Union[str, bytes, None] = None,
             time_cost: int = DEFAULT_TIME_COST,
             memory_cost: int = DEFAULT_MEMORY_COST,
             parallelism: int = DEFAULT_PARALLELISM,
             hash_len: int = DEFAULT_HASH_LENGTH,
             salt_len: int = DEFAULT_SALT_LENGTH) -> "Argon2Hasher":
        return cls(hashed=PasswordHasher(
            time_cost=time_cost,
            memory_cost=memory_cost,
            parallelism=parallelism,
            hash_len=hash_len,
            salt_len=salt_len,
            encoding="utf-8",
            type=Type.ID
        ).hash(password, salt=Salt.generate(salt, salt_len).value))



# coding:utf-8

from urllib.parse import urljoin

__project__ = "xpw"
__version__ = "0.1.0"
__description__ = "Securely manage user passwords and authentication."
__urlhome__ = "https://github.com/bondbox/xpw/"
__urlcode__ = __urlhome__
__urldocs__ = __urlhome__
__urlbugs__ = urljoin(__urlhome__, "issues")

# author
__author__ = "Mingzhe Zou"
__author_email__ = "zoumingzhe@outlook.com"



# coding:utf-8

from typing import Iterable
from typing import Optional
from typing import Set

from ldap3 import ALL
from ldap3 import Attribute
from ldap3 import Connection
from ldap3 import Entry
from ldap3 import Server


class LdapAuth:
    def __init__(self, server: Server, username: str, password: str):
        self.__server: Server = server
        self.__bind_dn: str = username
        self.__bind_pw: str = password

    @property
    def server(self) -> Server:
        return self.__server

    @property
    def bind(self) -> Connection:
        return self.connect(self.server, self.__bind_dn, self.__bind_pw)

    @classmethod
    def connect(cls, server: Server, username: str, password: str) -> Connection:  # noqa:E501
        return Connection(server, username, password, auto_bind=True)

    def search(self, base: str, filter: str, attrs: Iterable[str], key: str) -> Optional[Entry]:  # noqa:E501
        """search entry"""
        attributes: Set[str] = set(attrs)
        bind: Connection = self.bind
        bind.search(base, filter, attributes=attributes)
        for entry in bind.entries:
            for attr in attributes:
                attribute: Attribute = getattr(entry, attr)
                if key in attribute.values:
                    return entry
        return None

    def verify(self, username: str, password: str) -> bool:
        """verify password"""
        try:
            return True if self.connect(self.server, username, password).bind() else False  # noqa:E501
        except Exception:
            return False

    def signed(self, base: str, filter: str, attrs: Iterable[str], username: str, password: str) -> Optional[Entry]:  # noqa:E501
        """search user and verify password"""
        user: Optional[Entry] = self.search(base, filter, attrs, username)
        return user if user and self.verify(user.entry_dn, password) else None


class LdapInit:
    def __init__(self, host: str, port: Optional[int] = None, use_ssl: bool = False):  # noqa:E501
        self.__server: Server = Server(host=host, port=port, use_ssl=use_ssl, get_info=ALL)  # noqa:E501

    def bind(self, username: str, password: str) -> LdapAuth:
        return LdapAuth(self.__server, username, password)

    @classmethod
    def from_url(cls, url: str) -> "LdapInit":
        return cls(host=url)
